"""
A helper file to aid with discord.py scripting. Using this requires you to have discord.py installed.
These Classes will not allow users to preform any major actions on your guilds, the most major being banning members.
Use these instead of passing raw discord.py models to your users, as these could be used to obtain sensitive materials,
such as your token, or to shut down your bot.
"""
from typing import Union, Optional

import discord
from discord import utils
from discord.ext import commands

from viper import VP_NONE, VP_ArgumentError, PyVP_Model

__all__ = [
    "SafeAccessTextChannel",
    "SafeAccessContext",
    "SafeAccessUser",
    "SafeAccessGuild",
    "SafeAccessMember",
    "SafeAccessMessage"
]


class SafeAccessContext(PyVP_Model):
    def __init__(self, ctx: commands.Context):
        self._ctx = ctx  # can't access underscored attrs, makes this safe
        self.message = SafeAccessMessage(ctx.message)
        self.author = SafeAccessMember(ctx.author) if ctx.guild else SafeAccessUser(ctx.author)
        self.channel = SafeAccessTextChannel(ctx.channel)
        self.guild = SafeAccessGuild(ctx.guild) if ctx.guild else VP_NONE
        self.bot = SafeAccessMember(ctx.me) if ctx.guild else SafeAccessUser(ctx.me)
        self.content = ctx.message.content

    def __repr__(self):
        return "<Context message={0}>".format(self.message)

    async def send(self, message: str, embed: discord.Embed = None) -> Optional["SafeAccessMessage"]:
        try:
            msg = await self._ctx.send(message, embed=embed)
            return SafeAccessMessage(msg)
        except discord.HTTPException:
            return VP_NONE


class SafeAccessMessage(PyVP_Model):
    def __init__(self, message: discord.Message):
        self._msg = message
        self.channel = SafeAccessTextChannel(message.channel)
        self.guild = SafeAccessGuild(message.guild) if message.guild else VP_NONE
        self.author = SafeAccessMember(message.author) if isinstance(message.author,
                                                                     discord.Member) else SafeAccessUser(message.author)
        self.content = message.content
        self.clean_content = message.clean_content
        self.flags = message.flags
        self.jump_url = message.jump_url

    def __repr__(self):
        return "<Message channel={0} guild={1} author={2}>".format(self.channel, self.guild, self.author)


class SafeAccessTextChannel(PyVP_Model):
    def __init__(self, channel: discord.TextChannel):
        self._chn = channel
        self.guild = SafeAccessGuild(channel.guild) if channel.guild else VP_NONE
        self.id = channel.id
        self.nsfw = channel.is_nsfw()
        self.news = channel.is_news()
        self.topic = channel.topic
        self.mention = channel.mention

    def __repr__(self):
        return "<Channel id={0} nsfw={1} guild={2}>".format(self.id, self.nsfw, self.guild)

    async def send(self, message: str, embed: discord.Embed = None) -> Optional[SafeAccessMessage]:
        try:
            msg = await self._chn.send(message, embed=embed)
            return SafeAccessMessage(msg)
        except discord.HTTPException:
            return VP_NONE


class SafeAccessMember(PyVP_Model):
    def __init__(self, member: discord.Member):
        self._mem = member
        self.name = member.name
        self.id = member.id
        self.discriminator = member.discriminator
        self.mention = member.mention
        self.guild = SafeAccessGuild(member.guild)
        self.nick = member.nick or VP_NONE

    def __repr__(self):
        return "<Member name={0} id={1} guild={2}>".format(self.name, self.id, self.guild)

    async def ban(self, reason=None):
        await self._mem.ban(reason=reason)
        return VP_NONE

    async def unban(self, reason=None):
        await self._mem.unban(reason=reason)
        return VP_NONE

    async def send_dm(self, message, embed: discord.Embed = None) -> Optional[SafeAccessMessage]:
        try:
            msg = await self._mem.send(message, embed=embed)
            return SafeAccessMessage(msg)
        except discord.HTTPException:
            return VP_NONE


class SafeAccessUser(PyVP_Model):
    def __init__(self, user: discord.User):
        self._usr = user
        self.name = user.name
        self.id = user.id
        self.discriminator = user.discriminator
        self.mention = user.mention

    def __repr__(self):
        return "<User name={0} id={1}>".format(self.name, self.id)

    async def send_dm(self, message, embed: discord.Embed = None) -> Optional[SafeAccessMessage]:
        try:
            msg = await self._usr.send(message, embed=embed)
            return SafeAccessMessage(msg)
        except discord.HTTPException:
            return VP_NONE

class SafeAccessGuild(PyVP_Model):
    def __init__(self, guild: discord.Guild):
        self._guild = guild
        self.member_count = guild.member_count
        self.owner = SafeAccessUser(guild.owner) # this needs to be a user to avoid recursion
        self.description = guild.description
        self.name = guild.name
        self.id = guild.id

    def __repr__(self):
        return "<Guild name={0} id={1} owner={2}>".format(self.name, self.id, self.owner)

    def get_member(self, id_or_name: Union[str, int]) -> Optional[SafeAccessMember]:
        if isinstance(id_or_name, int):
            member = self._guild.get_member(id_or_name)
            if member:
                return SafeAccessMember(member)

            return VP_NONE

        elif isinstance(id_or_name, str):
            member = self._guild.get_member_named(id_or_name)
            if member:
                return SafeAccessMember(member)

            return VP_NONE

        raise VP_ArgumentError("get_member expected a string or an integer, not {0!r}".format(id_or_name))

    def get_channel(self, id_or_name: Union[str, int]) -> Optional[SafeAccessTextChannel]:
        if isinstance(id_or_name, int):
            channel = self._guild.get_channel(id_or_name)
            if channel:
                return SafeAccessTextChannel(channel)

            return VP_NONE

        elif isinstance(id_or_name, str):
            channel = utils.get(self._guild.text_channels, name=id_or_name)
            if channel:
                return SafeAccessTextChannel(channel)

            return VP_NONE

        raise VP_ArgumentError("get_channel expected a string or an integer, not {0!r}".format(id_or_name))



from .objects import VP_Error

class VP_RecursionError(VP_Error):
    pass

class VP_ExecutionError(VP_Error):
    pass

class VP_SyntaxError(VP_ExecutionError):
    pass

class VP_ArgumentError(VP_ExecutionError):
    pass

class VP_NameError(VP_ExecutionError):
    pass

class VP_StaticError(VP_ExecutionError):
    pass

class VP_AttributeError(VP_Error):
    pass

class VP_RaisedError(VP_Error):
    def __init__(self, message):
        self.message = message
        super(VP_RaisedError, self).__init__(message)

    def __repr__(self):
        return f"[Raised Error: {self.message}]"

    def __str__(self):
        return self.message


from .errors import *
from .keywords import *

class VPNamespace(dict):
    __buildmode = False
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self[VIPER_KEYWORDS['VIPER_TRUE']] = True, True
        self[VIPER_KEYWORDS['VIPER_FALSE']] = False, True

    def buildmode(self, state: bool):
        self.__buildmode = state

    def force_assign(self, key: str, value, static: bool=False):
        dict.__setitem__(self, key, (value, static))
    
    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default
    
    def __getitem__(self, key):
        v = dict.__getitem__(self, key)
        try:
            return v[0]
        except:
            return v
        
    def __setitem__(self, key, value):
        static = False
        if isinstance(value, tuple):
            static = value[1]
            value = value[0]
        
        if self.__buildmode:
            return dict.__setitem__(self, key, (value, static))
        
        exists = dict.get(self, key, None)
        if exists is None:
            dict.__setitem__(self, key, (value, static))
        
        else:
            if exists[1]:
                raise VP_StaticError(key)
            
            dict.__setitem__(self, key, (value, static))

def find_outer_brackets(
    raw_code: str,
    bracket_in: str = None,
    bracket_out: str = None,
    include_brackets: bool = True,
    skip_extra: bool = False,
    return_excess: bool = False
    ) -> str:
    """
    returns code up to the outermost bracket
    """
    bracket_in = bracket_in or VIPER_BRACKETS['VIPER_CODE_IN']
    bracket_out = bracket_out or VIPER_BRACKETS['VIPER_CODE_OUT']
    
    raw_code = raw_code.strip()
    iterator = enumerate(raw_code)
    output = ""
    level = 0
    
    if not raw_code.startswith(bracket_in):
        if not skip_extra:
            raise VP_SyntaxError("Start character, '{0}', not found".format(bracket_in))
        
        else:
            _, char = next(iterator)
            try:
                while char != bracket_in:
                    _, char = next(iterator)
                
                if include_brackets:
                    output += bracket_in
                level += 1
            except StopIteration:
                raise VP_SyntaxError("Start character, '{0}', not found".format(bracket_in))
    
    for index, char in iterator:
        if include_brackets:
            output += char
        
        if char == bracket_in:
            level += 1
            if level != 1:
                output += char
            
            continue
        
        if char == bracket_out:
            level -= 1
            if level == 0:
                if return_excess:
                    return raw_code[index+1:]

                return output
        
        if not include_brackets:
            output += char
    
    raise VP_SyntaxError("End character, '{0}', not found".format(bracket_out))



VP_NONE = None

class VPObject:
    __static = False
    def __init__(self, *args, **kwargs):
        self.init(*args, **kwargs)
    
    def init(self, *args, **kwargs):
        pass


class _VP_NULL(VPObject):
    __slots__ = tuple()
    def __new__(cls):
        global VP_NONE
        if VP_NONE is not None:
            return VP_NONE
        
        VP_NONE = object.__new__(cls)
        return VP_NONE
    
    def __str__(self):
        return "null"
    
    __repr__ = __str__

_VP_NULL()


class VP_Error(Exception):
    def __init__(self, *args, **kwargs):
        self.init(*args, **kwargs)
        super().__init__(*args)

    def init(self, msg=None):
        self.file = None
        self.line = None
        self.msg = msg
        self.stack = []
        self.message = msg if isinstance(msg, str) else repr(msg)

    def __str__(self):
        return self.format_stack()

    def format_stack(self):
        resp = ""
        for frame in reversed(self.stack):
            resp += f"\nFile {frame[2]}, {'in function '+frame[0].name if frame[0] else 'top-level'}:\n    {frame[1].strip()}"

        resp += "\n\n"

        resp += self.message

        return resp.strip()


class VP_LIST(VPObject):
    pass

class PyVP_Model:
    def __dir__(self):
        dirs = super(PyVP_Model, self).__dir__()
        dirs = [d for d in dirs if not d.startswith("_")] # hide attributes that cant be accessed
        return dirs


from .parser import build_code_async
from .objects import *
from .builtins import get_builtins
from .errors import *
from .common import *

__version__ = "0.0.1"

async def eval(raw_code, defaults=None, namespace=None, file="<string>", safe=False):
    """
    :param raw_code: :class:`str` the code to evaluate
    :param defaults: Optional[:class:`dict`] defaults to be statically injected into the namespace before parsing the file
    :param namespace: Optional[:class:`VPNamespace`] a pre-built namespace, useful to keep variables alive between evals
    :param file: :class:`str`
    :param safe: :class:`bool` whether "unsafe" builtins should be used. These include things like file-manipulation parameters.
    :return: :class:`VPNamespace`
    """
    if defaults is not None and not isinstance(defaults, dict):
        raise ValueError("defaults expected a dict")
    
    if namespace is not None and not isinstance(namespace, VPNamespace):
        raise ValueError("namespace expected a VPNamespace")
    
    module_ns = namespace or VPNamespace()
    module_ns.buildmode(True)

    module_ns.update(get_builtins(safe))

    if defaults:
        module_ns.update({a: ((b, True) if not isinstance(b, tuple) else b) for a, b in defaults.items()})
    
    module_ns['globals'] = module_ns
    module_ns.buildmode(False)
    
    await build_code_async(raw_code, module_ns, None, file=file)
    return module_ns

async def eval_file(fp, defaults=None, namespace=None, safe=False):
    """
    :param fp: :class:`str` the file path to read and parse
    :param defaults: Optional[:class`dict`] defaults to be statically injected into the namespace before parsing the file
    :param namespace: Optional[:class:`VPNamespace`] a pre-built namespace, useful to keep variables alive between evals
    :param safe: :class:`bool` whether "unsafe" builtins should be used. These include things like file-manipulation parameters.
    :return: :class:`VPNamespace`
    """
    with open(fp) as f:
        data = f.read()
    
    return await eval(data, defaults, namespace, file=fp, safe=safe)




import inspect

from .keywords import VIPER_BRACKETS, VIPER_STRICT_PARSING_KEYWORDS, VIPER_KEYWORDS
from .errors import *
from .objects import *
from .common import *
from . import functions

async def get_value_from_string(raw: str, global_ns, local_ns, depth):
    raw = raw.strip()

    if any([x in raw for x in "/*+-"]):
        return await build_math(raw, global_ns, local_ns, depth) # put this above the string parsing to allow string adding

    if raw.startswith('"'):
        # its a string
        resp = ""
        depth = False
        skip_one = False
        for index, char in enumerate(raw):
            if skip_one:
                skip_one = False
                continue
            
            if char == "\\":
                try:
                    resp += raw[index+1]
                    skip_one = True
                    continue
                except:
                    raise VP_SyntaxError("Invalid escape")
            
            if char == '"':
                if depth:
                    return resp
                depth = True
                continue
            if depth:
                resp += char

        raise VP_SyntaxError("Invalid string")

    
    if raw.startswith(VIPER_KEYWORDS['VIPER_VARMARKER']):
        if VIPER_BRACKETS['VIPER_FUNCTIONCALL_IN'] in raw:
            var = await call_function(raw, global_ns, local_ns, depth)
        else:
            var = get_variable(raw, global_ns, local_ns)

        return var
    
    if raw == VIPER_KEYWORDS['VIPER_TRUE']:
        return True
    
    if raw == VIPER_KEYWORDS['VIPER_FALSE']:
        return False
    
    if raw == VIPER_KEYWORDS['VIPER_NULL']:
        return VP_NONE
    
    if not raw.isalpha():
        try:
            return int(raw)
        except ValueError:
            try:
                return float(raw)
            except ValueError:
                pass
    
    if VIPER_BRACKETS['VIPER_FUNCTIONCALL_IN'] in raw:
        return await call_function(raw, global_ns, local_ns, depth)
    
    if local_ns is not None:
        var = local_ns.get(raw)
        if var is not None:
            return var
    
    var = global_ns.get(raw)
    if var is not None:
        return var
    
    raise VP_NameError(raw)

def get_variable_name(raw: str, marker=None):
    name = ""
    if raw in [VIPER_KEYWORDS['VIPER_TRUE'], VIPER_KEYWORDS['VIPER_FALSE']]:
        return raw
    
    marker = marker or VIPER_KEYWORDS['VIPER_VARMARKER']
    for index, char in enumerate(raw):
        if char == marker:
            for char_ in raw[index+1:]:
                if char_.isspace() or char == "(":
                    break
                name += char_
            break
    
    return name or None

def _get(name, global_ns, local_ns):
    resp = None
    
    if name == VIPER_KEYWORDS['VIPER_NULL']:
        return VP_NONE


    
    if local_ns is not None:
        resp = local_ns.get(name)
    
    if resp is None:
        resp = global_ns.get(name)
    
    return resp

async def build_math(raw: str, global_ns, local_ns=None, depth=None):
    args = []
    cur = ""
    for char in raw:

        if char not in ("+", "-", "*", "/"):
            cur += char
            continue

        if not cur:
            raise VP_SyntaxError("doubled math operators")

        if not args:
            args.append(cur)
            args.append([char])
            cur = ""
            continue

        else:
            args[-1].append(cur)
            args.append([char])
            cur = ""
            continue

    if cur:
        args[-1].append(cur)

    out = 0
    ch = {
        "/": lambda a, b: a / b,
        "*": lambda a, b: a * b,
        "+": lambda a, b: a + b,
        "-": lambda a, b: a - b
    }
    for index, arg in enumerate(args):
        if index is 0:
            continue
        if len(arg) != 2:
            raise VP_SyntaxError("something isnt right")
        front = args[index-1] if not isinstance(args[index-1], list) else out
        typ = arg[0]
        back = arg[1]
        if isinstance(front, str):
            front2 = await get_value_from_string(front, global_ns, local_ns, depth)
        else:
            front2 = front

        back = await get_value_from_string(back, global_ns, local_ns, depth)
        r = ch[typ](front2, back)
        if isinstance(r, int):
            if isinstance(front, str): # i really hate this part, maybe make math parsing not so bad later
                out += r
            else:
                out = r

        else:
            if isinstance(out, int):
                out = r
                continue

            out = r

    return out

def get_variable(raw: str, global_ns, local_ns=None, strict=True, marker=None):
    resp = _get(raw.strip(), global_ns, local_ns)
    if resp is not None:
        return resp
    
    name = get_variable_name(raw, marker)
    if name is None:
        if strict:
            raise VP_NameError("invalid variable name: '{0}'".format(raw))
        else:
            return None

    subs = name.split(".")
    if len(subs) > 1:
        cur = [subs[0]]
        resp = _get(subs[0], global_ns, local_ns)
        for attr in subs[1:]:
            cur.append(attr)

            if attr.startswith("_"):
                raise VP_AttributeError("underscored attributes are private, and cannot be accessed. ({0})".format(".".join(cur)))

            try:
                resp = getattr(resp, attr)
            except AttributeError:
                raise VP_AttributeError("{0} has no attribute {1} ({2})".format(resp, attr, ".".join(cur)))

    else:
        resp = _get(name, global_ns, local_ns)
    
    if resp is None and strict:
        raise VP_NameError("name '{0}' is not defined".format(name))
    
    return resp

async def get_variable_or_function_value(name: str, global_ns, local_ns, depth):
    if VIPER_BRACKETS['VIPER_FUNCTIONCALL_IN'] in name:
        resp = await call_function(name, global_ns, local_ns, depth)
    
    else:
        resp = await get_value_from_string(name, global_ns, local_ns, depth)
    
    return resp


async def call_function(line, global_ns, local_ns, depth):
    func = get_variable(line.split(VIPER_BRACKETS['VIPER_FUNCTIONCALL_IN'])[0].split("=")[-1].strip(), global_ns, local_ns, marker="")
    t = VIPER_BRACKETS['VIPER_FUNCTIONCALL_IN'] + VIPER_BRACKETS['VIPER_FUNCTIONCALL_IN'].join(line.split(VIPER_BRACKETS['VIPER_FUNCTIONCALL_IN'])[1:])
    args = await parse_call_arguments(t, global_ns, local_ns, depth)

    if isinstance(func, functions.VPFunction):
        resp = await func.Call(*args, depth=depth)
    elif inspect.iscoroutinefunction(func):
        resp = await func(*args)
    elif callable(func):
        resp = func(*args)
    else:
        raise VP_Error("{0} is not callable".format(func))

    return resp


async def parse_call_arguments(args: str, global_ns, local_ns, depth) -> list:
    raw_code = find_outer_brackets(args, VIPER_BRACKETS['VIPER_FUNCTIONCALL_IN'], VIPER_BRACKETS['VIPER_FUNCTIONCALL_OUT'], include_brackets=False)
    
    args = []
    argname = ""
    string = False
    argument_isstring = False
    bracketed = 0
    for index, char in enumerate(raw_code):
        if char.isspace():
            if string or bracketed > 0:
                argname += char

            continue

        if char == VIPER_BRACKETS['VIPER_FUNCTIONCALL_IN'] and not string:
            bracketed += 1

        if char == VIPER_BRACKETS['VIPER_FUNCTIONCALL_OUT'] and not string:
            bracketed -= 1

        if char == '"' and bracketed < 1:
            if string and (raw_code[index-1] if index is not 0 else "") != "\\":
                string = False
                continue

            elif not string:
                string = True
                argument_isstring = True

                continue
        
        if char == "," and not string and bracketed == 0:
            if not argname:
                raise VP_SyntaxError("Invalid comma")

            if argument_isstring:
                argname = '"' + argname + '"'
            args.append(argname)
            argument_isstring = False
            argname = ""
            continue
        
        argname += char

    if string:
        raise VP_SyntaxError("expected a closing string quote")

    if argname:
        if argument_isstring:
            argname = '"' + argname + '"'

        args.append(argname)

    resp = []
    for arg in args:
        v = await get_value_from_string(arg, global_ns, local_ns, depth)
        resp.append(v)
    
    return resp


async def viper_parse_return(line, global_ns, local_ns, depth):
    line = line.strip().replace(VIPER_KEYWORDS['VIPER_RETURN'], "", 1).strip()
    if not line:
        return VP_NONE
    
    resp = await get_value_from_string(line, global_ns, local_ns, depth)
    return resp or VP_NONE


async def compare_conditional(global_ns, local_ns, depth, arg1, operator=None, arg2=None):
    if not operator and not arg2:
        arg1 = await get_variable_or_function_value(arg1, global_ns, local_ns, depth)
        return bool(arg1)
    
    if operator and not arg2:
        raise VP_SyntaxError("invalid conditional (operator with no second argument)")
    
    arg1 = await get_variable_or_function_value(arg1, global_ns, local_ns, depth)
    arg2 = await get_variable_or_function_value(arg2, global_ns, local_ns, depth)
    
    if operator in VIPER_KEYWORDS['VIPER_EQUALS']:
        return arg1 == arg2
    
    if operator in VIPER_KEYWORDS['VIPER_NOT_EQUALS']:
        return arg1 != arg2
    
    if operator == VIPER_KEYWORDS['VIPER_GREATER']:
        return arg1 > arg2
    
    if operator == VIPER_KEYWORDS['VIPER_SMALLER']:
        return arg1 < arg2
    
    if operator == VIPER_KEYWORDS['VIPER_CONTAINS']:
        return arg1 in arg2

    if operator == VIPER_KEYWORDS['VIPER_NOTCONTAINS']:
        return arg1 not in arg2
    
    raise VP_ExecutionError("unknown operator")
    

async def parse_conditional(raw, global_ns, local_ns, depth):
    tests = find_outer_brackets(raw, VIPER_BRACKETS['VIPER_CONDITIONAL_IN'], VIPER_BRACKETS['VIPER_CONDITIONAL_OUT'], skip_extra=True, include_brackets=False)
    
    tests = tests.split()
    if len(tests) == 3 or len(tests) == 1:
        return await compare_conditional(global_ns, local_ns, depth, *tests)
    
    else:
        raise VP_SyntaxError("Invalid conditional (unknown number of conditionals, check there are not spaces in function calls)")

async def _build_line(raw_code, line, lineno, global_ns, local_ns, local, file, skipto, depth, func):
    if skipto is not None:
        if isinstance(skipto, int):
            if lineno < skipto:
                return skipto
            else:
                local['skipto'] = None
        else:
            if line.strip() == skipto.strip():
                local['skipto'] = None
            else:
                return
    
    line = line.strip()
    if line.startswith(VIPER_KEYWORDS['VIPER_COMMENT']) or not line:
        return
    
    static = False
    if line.startswith(VIPER_KEYWORDS['VIPER_STATIC']):
        static = True
        line = line.replace(VIPER_KEYWORDS['VIPER_STATIC'], "", 1).strip()
    
    if line.startswith(VIPER_KEYWORDS['VIPER_FUNC']):
        if local_ns is not None:
            raise VP_SyntaxError("Nested functions are not allowed")
        
        code = line.join(raw_code.split(line)[1:])
        code = line + code
        
        func = functions.VPFunction(code, file, global_ns)
        global_ns[func.name] = func, static
        return lineno + len(func.code.splitlines())

    if line.startswith(VIPER_KEYWORDS['VIPER_VARMARKER']):
        name = get_variable_name(line)
        if name is None:
            raise VP_SyntaxError("invalid variable name")

        getter = line.split("=")

        if len(getter) < 2:
            if VIPER_BRACKETS['VIPER_FUNCTIONCALL_IN'] in line:
                await call_function(line, global_ns, local_ns, depth)
                return

            raise VP_SyntaxError("declaring variable must have an assignment")

        if VIPER_BRACKETS['VIPER_FUNCTIONCALL_IN'] in line:
            value = await call_function(line, global_ns, local_ns, depth)
        
        else:
            value = await get_value_from_string(getter[1], global_ns, local_ns, depth)
        
        if local_ns is not None:
            local_ns[name] = value, static
        
        else:
            global_ns[name] = value, static
        
        return
    
    if line.startswith(VIPER_KEYWORDS['VIPER_RETURN']):
        if local_ns is None:
            raise VP_SyntaxError("return outside of function")
        
        resp = viper_parse_return(line, global_ns, local_ns, depth)
        err = StopIteration(resp)
        raise err

    if line.startswith(VIPER_KEYWORDS['VIPER_THROW']):
        value = await get_value_from_string(line.replace(VIPER_KEYWORDS['VIPER_THROW'], "", 1).strip(), global_ns, local_ns, depth)
        raise VP_RaisedError(repr(value))
    
    if line.startswith(VIPER_KEYWORDS['VIPER_IF']):
        c = "\n".join(raw_code.splitlines()[lineno:])
        conditional_code = find_outer_brackets(c, skip_extra=True, include_brackets=False)
        _cond = find_outer_brackets(c, skip_extra=True)
        excess = find_outer_brackets(c, skip_extra=True, return_excess=True)
        length = len(_cond.splitlines())
        
        found_one = False
        if await parse_conditional(line, global_ns, local_ns, depth):
            found_one = True
            maybe_resp = await build_code_async(conditional_code, global_ns, local_ns, file=file, depth=depth, func=func)
            if maybe_resp is not None:
                raise StopIteration(maybe_resp)
        
        c = excess.strip()

        while c.startswith(VIPER_KEYWORDS['VIPER_ELIF']):
            conditional_code = find_outer_brackets(c, skip_extra=True, include_brackets=False)
            _cond = find_outer_brackets(c, skip_extra=True)
            excess = find_outer_brackets(c, skip_extra=True, return_excess=True)
            length += len(_cond.splitlines())
            
            if not found_one and await parse_conditional(c.splitlines()[0], global_ns, local_ns, depth):
                found_one = True
                maybe_resp = await build_code_async(conditional_code, global_ns, local_ns, file=file, depth=depth, func=func)
                if maybe_resp is not None:
                    raise StopIteration(maybe_resp)

            c = excess.strip()


        if c.startswith(VIPER_KEYWORDS['VIPER_ELSE']):
            v = c.replace(VIPER_KEYWORDS['VIPER_ELSE'], "", 1).strip()
            conditional_code = find_outer_brackets(v, skip_extra=True, include_brackets=False)
            _cond = find_outer_brackets(v, skip_extra=True)
            length += len(_cond.splitlines())

            if not found_one:
                maybe_resp = await build_code_async(conditional_code, global_ns, local_ns, file=file, depth=depth, func=func)
                if maybe_resp is not None:
                    raise StopIteration(maybe_resp)

        return length + lineno

    if line.startswith(VIPER_KEYWORDS['VIPER_TRY']):
        c = "\n".join(raw_code.splitlines()[lineno:])
        maybe_error = find_outer_brackets(c, skip_extra=True, include_brackets=False)
        maybe_error_code = find_outer_brackets(c, skip_extra=True)
        rest = find_outer_brackets(c, skip_extra=True, return_excess=True).strip()
        length = len(maybe_error_code.splitlines())
        find_end = lineno + length
        while not raw_code.splitlines()[find_end].strip():
            find_end += 1

        if not rest.startswith(VIPER_KEYWORDS['VIPER_CATCH']):
            raise VP_SyntaxError(f"\"{VIPER_KEYWORDS['VIPER_TRY']}\" block without a \"{VIPER_KEYWORDS['VIPER_CATCH']}\" statement")

        catch_block = find_outer_brackets(rest, skip_extra=True, include_brackets=False).strip()
        catch_end = len(find_outer_brackets(rest, skip_extra=True, include_brackets=True).splitlines()) + find_end

        try:
            await build_code_async(maybe_error, global_ns, local_ns, file=file, depth=depth, func=func)
        except VP_Error as error:
            if local_ns:
                local_ns.force_assign("error", error, True)
            else:
                global_ns.force_assign("error", error, True)

            await build_code_async(catch_block, global_ns, local_ns, file=file, depth=depth, func=func)

            if local_ns:
                del local_ns['error']

            else:
                del global_ns['error']

        return catch_end

    if VIPER_BRACKETS['VIPER_FUNCTIONCALL_IN'] in line:
        await call_function(line, global_ns, local_ns, depth)
        return
    
    raise VP_SyntaxError("something isnt right: {0}".format(line))

async def build_code_async(raw_code: str, global_namespace, local_namespace=None, file="<input>", depth=0, func=None):
    """
    allows for auto-awaiting of async functions
    :param raw_code: the code to evaluate
    :param global_namespace: the global VP_Namespace
    :param local_namespace: the local VP_Namespace, if in a function context
    :param file: the file we are currently in
    :param depth: the parse depth, used to prevent python recursionerrors
    :param func: the function we are currently in, used to gather stack traces
    :return: Optional[Any]
    """
    if depth >= 100:
        raise VP_RecursionError("max depth while parsing viper")  # cutting it close to pythons default recursion depth

    lines = raw_code.splitlines()
    skipto = None

    for index, line in enumerate(lines):
        if skipto and index <= skipto:
            continue

        skipto = None
        try:
            r = await _build_line(raw_code, line, index, global_namespace, local_namespace, locals(), file, skipto, depth, func)
            if r is not None:
                skipto = r
        except StopIteration as e:
            return e.args[0]

        except VP_Error as error:
            error.file = file
            error.line = line
            error.stack.append((func, line, file))

            raise



import random
import os
from typing import Union, Any

from .objects import VP_NONE, VP_Error
from .errors import VP_ArgumentError


def read(filepath: str) -> Union[str, type(VP_NONE)]:
    """
    reads a file and returns the contents (not a byte file)
    :param filepath: :class:`str` the filepath to read
    :return: :class:`str` the file contents
    """
    if os.path.exists(filepath):
        with open(filepath, "r") as f:
            return f.read()

    else:
        return VP_NONE


def randline(filepath: str) -> Union[str, type(VP_NONE)]:
    """
    Reads a file and returns a random line
    :param filepath: :class:`str` the filepath to read
    :return: :class:`str` or :class:`VP_NONE` if the file is empty
    """
    if os.path.exists(filepath):
        with open(filepath) as f:
            lines = f.readlines()
            if not lines:
                return VP_NONE

            return random.choice(lines)

    else:
        return VP_NONE


def write(filepath: str, content: str) -> type(VP_NONE):
    """
    writes the content to the given filepath
    :param filepath: the file to write to
    :param content:
    :return: :class:`VP_NONE`
    """
    if not os.path.exists(os.path.dirname(filepath)):
        os.makedirs(os.path.dirname(filepath))

    with open(filepath, "w") as f:
        f.write(content)

    return VP_NONE

def add_line(filepath: str, content: str) -> type(VP_NONE):
    """
    adds a line to a pre-existing file
    :param filepath: :class:`str` the file path to write to
    :param content: :class:`str` the content to write
    :return: :class:`VP_NONE`
    """
    if not os.path.exists(os.path.dirname(filepath)):
        os.makedirs(os.path.dirname(filepath))

    with open(filepath, "a") as f:
        f.write(content)

    return VP_NONE

def randnum(min: int, max: int) -> int:
    """
    returns a random number between the min and max
    :param min: :class:`int` the minimum number to return
    :param max: :class:`int` the maximum number to return
    :return: :class:`int` the chosen number
    """
    return random.randint(min, max)

def choice(*choices: Any) -> Any:
    """
    returns a random choice from the given input. takes any number of arguments.
    :param choices: the choices to pick from
    :return: Any. the chosen item from the input arguments. returns :class:`VP_NONE` if no arguments given
    """
    if not choices:
        return VP_NONE
    
    return random.choice(choices)

def say(*args):
    """
    prints the given text to the console.
    :param args:
    :return:
    """
    print(*args)
    return VP_NONE

def to_num(s):
    try:
        return int(s)
    except:
        return float(s)

def stack(error: VP_Error):
    """
    utility for getting a traceback from an error
    :param error: the error
    :return: str
    """
    if not isinstance(error, VP_Error):
        raise VP_ArgumentError("expected an error, got {0!r}".format(error))

    return error.format_stack()

def fmt(string: str, *args, **kwargs):
    return string.format(*args, **kwargs)


safe_builtins = {
    "say": say,
    "string": str,
    "boolean": bool,
    "number": to_num,
    "choice": choice,
    "randnum": randnum,
    "stack": stack,
    "format": fmt
}

unsafe_builtins = {
    "read": read,
    "write": write,
    "writeline": add_line,
    "readrandline": randline,
    **safe_builtins
}

def get_builtins(safety=False):
    if safety:
        return {a: (b, False) for a, b in safe_builtins.items()}

    return {a: (b, False) for a, b in unsafe_builtins.items()}




from .objects import VPObject, VP_NONE
from .common import VPNamespace, find_outer_brackets
from .errors import VP_SyntaxError, VP_ArgumentError
from .keywords import VIPER_KEYWORDS, VIPER_BRACKETS

def interpret_arguments(raw_code: str):
    raw_code = find_outer_brackets(raw_code, VIPER_BRACKETS['VIPER_FUNCTIONARGS_IN'], VIPER_BRACKETS['VIPER_FUNCTIONARGS_OUT'], include_brackets=False)
    args = []
    
    argname = ""
    optional = False
    for char in raw_code:
        if char.isspace():
            continue
        
        if char == ",":
            if not argname:
                raise VP_SyntaxError("Invalid comma")
            
            args.append((argname, optional))
            argname = ""
            optional = False
            continue
        
        if char == "?" and not argname:
            optional = True
            continue
        
        elif char == "?" and argname:
            raise VP_SyntaxError("Invalid optional character, {0}".format(argname))
        
        argname += char
    
    if argname:
        args.append((argname, optional))
    
    return args

def get_function_name(raw_code):
    raw_code = raw_code.strip()
    name = ""
    for char in raw_code:
        if char == VIPER_BRACKETS['VIPER_FUNCTIONARGS_IN']:
            break
        
        name += char
    
    if not name:
        raise VP_SyntaxError("No function name")
    
    return name.strip()

class VPFunction(VPObject):
    __slots__ = ("raw_code", "code", "_code", "file", "global_ns")
    def __init__(self, raw_code: str, fp: str, namespace: VPNamespace):
        self.raw_code = raw_code
        code = raw_code.replace(VIPER_KEYWORDS['VIPER_FUNC'], "", 1)
        code = code.strip()
        self.name = get_function_name(code)
        code = code.replace(self.name, "", 1).strip()
        self.file = fp
        self.arguments = interpret_arguments(code)
        self.code = find_outer_brackets(code, skip_extra=True)
        self._code = find_outer_brackets(code, skip_extra=True, include_brackets=False)
        self.global_ns = namespace
        self.init()

    def __repr__(self):
        return "<Function {0} in file {1}>".format(self.name, self.file)
    
    async def Call(self, *args, depth=None):
        args = list(reversed(args))
        if depth is None:
            raise ValueError("depth is a required argument")
        
        ns = VPNamespace()
        ns.buildmode(True)
        for index, (name, optional) in enumerate(self.arguments):
            try:
                ns[name] = args.pop()
            except:
                if optional:
                    ns[name] = VP_NONE
                    break
                raise VP_ArgumentError("Missing required argument '{0}' for function call".format(name))

        if args:
            raise VP_ArgumentError("Too many arguments passed to " + self.name)
        
        ns.buildmode(False)
        resp = await parser.build_code_async(self._code, self.global_ns, ns, self.file, depth+1, func=self)
        return resp

from . import parser



VIPER_KEYWORDS = {
    "VIPER_STATIC": "static",
    "VIPER_FUNC": "func",
    "VIPER_NULL": "None",
    
    "VIPER_IF": "if",
    "VIPER_ELIF": "else if",
    "VIPER_ELSE": "else",
    
    "VIPER_FALSE": "false",
    "VIPER_TRUE": "true",
    
    "VIPER_NAMESPACE": "$namespace",
    "VIPER_RETURN": "return",
    
    "VIPER_EQUALS": ["is", "=="],
    "VIPER_NOT_EQUALS": ["is not", "!="],
    "VIPER_GREATER": ">",
    "VIPER_SMALLER": "<",
    "VIPER_CONTAINS": "in",
    "VIPER_NOTCONTAINS": "!in",
    
    "VIPER_VARMARKER": "$",
    
    "VIPER_COMMENT": "//",

    "VIPER_TRY": "try",
    "VIPER_CATCH": "error",
    "VIPER_THROW": "raise"
}

# don't touch this one
VIPER_STRICT_PARSING_KEYWORDS = [
    VIPER_KEYWORDS['VIPER_FUNC'],
    VIPER_KEYWORDS['VIPER_STATIC'],
    VIPER_KEYWORDS['VIPER_RETURN']
]

VIPER_BRACKETS = {
    "VIPER_CODE_IN": "{",
    "VIPER_CODE_OUT": "}",
    "VIPER_FUNCTIONARGS_IN": "(",
    "VIPER_FUNCTIONARGS_OUT": ")",
    "VIPER_FUNCTIONCALL_IN": "(",
    "VIPER_FUNCTIONCALL_OUT": ")",
    "VIPER_CONDITIONAL_IN": "(",
    "VIPER_CONDITIONAL_OUT": ")"
}
